#!/usr/bin/env python

# A native Python implementation of the SWORD Project Bible Reader
# Currently only ztext Bible modules are implemented.

# * ztext format documentation
# I'll use Python's struct module's format strings.
# See http://docs.python.org/lib/module-struct.html
# Take the Old Testament (OT) for example. Three files:
#
#  - ot.bzv: Maps verses to character ranges in compressed buffers.
#    10 bytes ('<IIH') for each verse in the Bible:
#    - buffer_num (I): which compressed buffer the verse is located in
#    - verse_start (I): the location in the uncompressed buffer where the verse begins
#    - verse_len (H): length of the verse, in uncompressed characters
#    These 10-byte records are densely packed, indexed by VerseKey 'Indicies' (docs later).
#    So the record for the verse with index x starts at byte 10*x.
#
#  - ot.bzs: Tells where the compressed buffers start and end.
#    12 bytes ('<III') for each compressed buffer:
#    - offset (I): where the compressed buffer starts in the file
#    - size (I): the length of the compressed data, in bytes
#    - uc_size (I): the length of the uncompressed data, in bytes (unused)
#    These 12-byte records are densely packed, indexed by buffer_num (see previous).
#    So the record for compressed buffer buffer_num starts at byte 12*buffer_num.
#
#  - ot.bzz: Contains the compressed text. Read 'size' bytes starting at 'offset'.
#
#  NT is analogous.
#
# Example usage:
#  python pysword.py esv 1pet 2 9

import os
import string
import data
from books import ref_to_index, testaments, find_book, Book

import struct, zlib
from os.path import join as path_join

class ZModule(object):
    def __init__(self, module):
        self.module = module
        try:
	    self.files = {
		'ot': self.get_files('ot'),
		'nt': self.get_files('nt')
		}
	    #return 0
	except:
	    print "init failed"
	    self.files = {
		'ot': [""],
		'nt': [""]
		}
	    #return None
   
    def get_files(self, testament):
        '''Given a testament ('ot' or 'nt'), returns a tuple of files
        (verse_to_buf, buf_to_loc, text)
        '''
        base = path_join(data.modules_path, self.module)
        v2b_name, b2l_name, text_name = [path_join(base, '%s.bz%s' % (testament, code))
                                         for code in ('v', 's', 'z')]
        #try:
	return [open(name, 'rb') for name in (v2b_name, b2l_name, text_name)]
	#except:
	    #print "Failed to get file"
	    
	    #return -1


    def text_for_index(self, testament, index):
        '''Get the text for a given index.'''
        try:
	    verse_to_buf, buf_to_loc, text = self.files[testament]
	except:
	    print "text_for_index failed"
	    return ""

        # Read the verse record.
        verse_to_buf.seek(10*index)
        buf_num, verse_start, verse_len = struct.unpack('<IIH', verse_to_buf.read(10))
       
        uncompressed_text = self.uncompressed_text(testament, buf_num)
        return uncompressed_text[verse_start:verse_start+verse_len]



    def uncompressed_text(self, testament, buf_num):
        verse_to_buf, buf_to_loc, text = self.files[testament]

        # Determine where the compressed data starts and ends.
        buf_to_loc.seek(buf_num*12)
        offset, size, uc_size = struct.unpack('<III', buf_to_loc.read(12))

        # Get the compressed data.
        text.seek(offset)
        compressed_data = text.read(size)
        return zlib.decompress(compressed_data)
        
        

    def text_for_ref(self, book, chapter, verse):
        '''Get the text for a given reference'''
        chapter, verse = int(chapter), int(verse)
        testament, idx = ref_to_index(book, chapter, verse)
        text = self.text_for_index(testament, idx)
        text = remove_all_tags(text)
        return text



    def all_verses_in_testament(self, testament):
        books = testaments[testament]
        for book in books:
            for verse in self.all_verses_in_book(book):
                yield verse



    def all_verses_in_book(self, book):
        if not isinstance(book, Book): book = find_book(book)
        for chapter, verses in enumerate(book.chapter_lengths):
            chapter = chapter + 1
            for verse in range(verses):
                verse = verse+1
                yield (book, chapter, verse,
                       self.text_for_index(book.testament,
                                           book.get_index_for_ref(chapter, verse)))
            


    def name_of_book(self, book):
      	arr = find_book(book)
	#print "name of book", book, "is", arr.name
	return arr.name
      


    def number_of_chapters_in_book(self, book):
	arr = find_book(book)
	chapters = arr.num_chapters
	print book, "has", chapters, "chapters"
	return chapters




    def number_of_verses_in_chapter(self, book, chapter):
	testament, idx = ref_to_index(book, int(chapter), 1)
	arr= find_book(book)
	verses = arr.chapter_lengths[int(chapter)-1]
	print book, chapter, "has", verses, "verses"
	return verses
	
	

	



def remove_tag(text, search_tag_start, search_tag_end):
    i=0
    #try:
    #search_tag_start="<note"
    #search_tag_end="/note>"
    reached_end=False
    while(reached_end==False):
	startpos=string.find(text,search_tag_start)
	if(startpos<>-1): #found tag
	    endpos=string.find(text,search_tag_end,startpos)
	    tag=text[startpos:endpos+len(search_tag_end)]
	    #print startpos, endpos, tag
	    text=string.replace(text, tag,"")
	else: #no more tags
	  reached_end=True
	      

    #except:
        #print "parsing of tag <note> failed (line="+str(i)+")"
  
    return text


def remove_all_tags(text):
    #ToDo: improve filter, remove forgotten tags, instead of removing tags like title, show them in a nicly formated way.
    text=remove_tag(text, "<note", "/note>")
    text=remove_tag(text, "<title", "/title>")
    text=remove_tag(text, "<lb", ">")
    text=remove_tag(text, "<milestone", ">")
    text=remove_tag(text, "<lg", ">")
    text=remove_tag(text, "<l", ">")
    text=remove_tag(text, "<q", ">")
    return text





#if __name__=='__main__':
    #import sys
    #mod_name, book, chapter, verse = sys.argv[1:]
   
    #module = ZModule(mod_name)
    
    #verses=module.number_of_verses_in_chapter(book, chapter)
    
    #for i in range(1, verses+1):
      #text = module.text_for_ref(book, chapter, i)
      #text = remove_all_tags(text)
      ##print "---------------"
      #print book, chapter, i, "   ", text 
    
    
    
    
    
    
    
    
    
    
